#include "filetree.h"
#include "droptarget.h"
#include "uthread.h"

enum
{
	O_TREE = wxID_HIGHEST,
};

BEGIN_EVENT_TABLE(UFiletree, wxTreeCtrl)
	EVT_TREE_BEGIN_DRAG(O_TREE, UFiletree::OnDragBegin)
	EVT_TREE_END_DRAG(O_TREE, UFiletree::OnDragEnd)
	EVT_TREE_ITEM_EXPANDING(O_TREE, UFiletree::OnItemExpanding)
	EVT_TREE_ITEM_COLLAPSED(O_TREE, UFiletree::OnItemCollapsed)
	EVT_TREE_ITEM_ACTIVATED(O_TREE, UFiletree::OnItemActivated)
	EVT_TREE_BEGIN_LABEL_EDIT(O_TREE, UFiletree::OnItemBeginRename)
	EVT_TREE_END_LABEL_EDIT(O_TREE, UFiletree::OnItemEndRename)
	EVT_LEAVE_WINDOW(UFiletree::OnLeaveWindow)
END_EVENT_TABLE()

UFiletree::UFiletree(wxWindow *oParent, wxWindowID iID, wxTextCtrl *oLocation, MainWindow *oParentWindow, Settings *oNewSettings)
: wxTreeCtrl(oParent, O_TREE, wxDefaultPosition, wxDefaultSize, wxTR_HAS_BUTTONS | wxTR_TWIST_BUTTONS | wxTR_DEFAULT_STYLE | wxTR_ROW_LINES | wxTR_MULTIPLE)
{
	bSeeHidden = false;
	oSettings = oNewSettings;
	txtLocation = oLocation;
	oParentFrame = oParentWindow;
	SetIndent(14);
	SetSpacing(8);
	SetBackgroundColour(wxColour(240, 240, 240));
	SetDropTarget(new UDropTarget(this));
	selectTreeItem(GetRootItem());
	
	createThread();
}

void UFiletree::dropFiles(std::vector<wxString> sFiles, wxCoord oX, wxCoord oY)
{
	wxTreeItemId iItem = HitTest(wxPoint(oX, oY));
	wxString sDestination = getFullpath(iItem);
	if (UFileHandler::isFile((const char *)sDestination.mb_str()))
	{
		sDestination = sDestination.Mid(0, sDestination.Find('/', true));
	}
	if (UFileHandler::isDir((const char *)sDestination.mb_str()))
	{
		for (unsigned int i = 0; i < sFiles.size(); i++)
		{
			UFileHandler::move((const char *)sFiles[i].mb_str(), (const char *)sDestination.mb_str());
		}
	}
}

void UFiletree::OnLeaveWindow(wxMouseEvent &oEvent)
{
	if (oEvent.LeftIsDown())
	{
		wxFileDataObject oFiles;
		
		wxArrayTreeItemIds iSelected;
		GetSelections(iSelected);
		for (unsigned int i = 0; i < iSelected.GetCount(); i++)
		{
			oFiles.AddFile(getFullpath(iSelected[i]));
		}
		
		oDropSource = new UDropSource(this);
		oDropSource->SetData(oFiles);
		
		//printf("Dragging something off the table!\n");
		wxString sString;
		switch ( oDropSource->DoDragDrop(wxDrag_DefaultMove) )
		{
			case wxDragError:   sString = _T("Error!");    break;
			case wxDragNone:    sString = _T("Nothing");   break;
			case wxDragCopy:    sString = _T("Copied");    break;
			case wxDragMove:    sString = _T("Moved");     break;
			case wxDragCancel:  sString = _T("Cancelled"); break;
			default:            sString = _T("Huh?");      break;
	        }
		wxSetCursor(0);
		//printf("!%s\n", sString.mb_str());
	}
}

void UFiletree::createThread()
{
	oThread = new UThread(this);
	oThread->Create();
	oThread->Run();
}

wxString UFiletree::getFullpath(wxTreeItemId iItem)
{
	if (!itemExists(iItem))
	{
		return wxString(_("!"));
	}
	wxTreeItemId iNewID = iItem;
	wxString sDir;
	sDir = GetItemText(iItem);
	while (iNewID != GetRootItem())
	{
		iNewID = GetItemParent(iNewID);
		sDir = GetItemText(iNewID) + _("/") + sDir;
	}
	sDir.Replace(_("//"), _("/"));
	return sDir;
}

bool UFiletree::deleteItem()
{
	if (!somethingSelected())
	{
		return false;
	}
	
	std::vector<wxTreeItemId> iItems = getTreeSelections();
	
	// Security measures.
	// Also, what are we going to view if we delete everything we see?
	// Going to the upper level directory would probably do nothing but confuse the user.
	// (Besides, if I'd accidentally do such a stupid thing and the program was arrogant enough to list that as a feature... I'd strangle a puppy.)
	for (unsigned int i = 0; i < iItems.size(); i++)
	{
		if (GetRootItem() == iItems[i])
		{
			return false;
		}
	}
	
	wxMessageDialog oDialog(this, _("Are you sure you want to delete these items?"), _("Delete items"), wxYES_NO);
	int iResult = oDialog.ShowModal();
	if (iResult == wxID_YES)
	{
		unsigned int iOKs = 0;
		for (unsigned int i = 0; i < iItems.size(); i++)
		{
			if (!itemExists(iItems[i]))
			{
				continue;
			}
			wxTreeItemId iParent = GetItemParent(iItems[i]);
			wxString sItem = getFullpath(iItems[i]);
			wxString sRunString;
			if (wxDir::Exists(sItem))
			{
				sRunString = _("rm -rf \"");
				oTimeMonitor.delTime((const char *)sItem.mb_str());
			}
			else
			{
				sRunString = _("rm -f \"");
			}
			sRunString += sItem + _("\"");
			iResult = system((const char *)sRunString.mb_str());
			if (iResult == 0)
			{
				wxTreeItemId iNext;
				wxTreeItemId iPrev;
				wxString sParentDir;
				sParentDir = getFullpath(iParent);
				
				if (itemExists(iItems[i]))
				{
					iNext = GetNextSibling(iItems[i]);
					iPrev = GetPrevSibling(iItems[i]);
					
					if (itemExists(iNext))
					{
						//selectTreeItem(iNext);
					}
					else if (itemExists(iPrev))
					{
						selectTreeItem(iPrev);
					}
					else
					{
						selectTreeItem(iParent);
					}
				}
				
				wxDir oDir(sParentDir);
				if (!oDir.HasFiles() && !oDir.HasSubDirs())
				{
					Collapse(iParent);
					SetItemHasChildren(iParent, false);
				}
				iOKs++;
			}
		}
		if (iOKs == iItems.size())
		{
			return true;
		}
	}
	return false;
}

void UFiletree::selectPrevItem()
{
	selectTreeItem(getPrevID(getSingleTreeSelection()));
}

void UFiletree::OnDragBegin(wxTreeEvent &oEvent)
{
	iMovingID = oEvent.GetItem();
	oEvent.Allow();
}

void UFiletree::OnDragEnd(wxTreeEvent &oEvent)
{
	wxString sOldItem;
	wxString sDestination;
	wxString sRunString;
	if (wxDir::Exists(getFullpath(oEvent.GetItem())))
	{
		sOldItem = getFullpath(this->iMovingID);
		sDestination = getFullpath(oEvent.GetItem());
		if (sOldItem == _("!"))
		{
			return;
		}
		else if (sOldItem == sDestination + _("/") + GetItemText(this->iMovingID))
		{
			return;
		}
		UFileHandler::move((const char *)sOldItem.mb_str(), (const char *)sDestination.mb_str());
	}
}

wxTreeItemId UFiletree::addItem(wxTreeItemId iDirID, const wxString &sItem)
{
	wxTreeItemId iTempID;
	wxString sFullpath(getFullpath(iDirID) + _("/") + sItem);
	int iImage;
	if (wxDir::Exists(sFullpath))
	{
		if (wxFile::Access(sFullpath, wxFile::read))
		{
			iTempID = AppendItem(iDirID, sItem, (UFileHandler::isSymlink((const char *)sFullpath.mb_str()) ? 6 : 1));
			wxDir oDir(sFullpath);
			if (oDir.HasFiles() || oDir.HasSubDirs())
			{
				SetItemHasChildren(iTempID, true);
			}
		}
		else
		{
			iTempID = AppendItem(iDirID, sItem, 3);
		}
	}
	else
	{
		if (UFileHandler::isSymlink((const char *)sFullpath.mb_str()))
		{
			iImage = (UFileHandler::isExecutable((const char *)sFullpath.mb_str()) ? 5 : 4);
		}
		else
		{
			iImage = (UFileHandler::isExecutable((const char *)sFullpath.mb_str()) ? 2 : 0);
		}
		
		iTempID = AppendItem(iDirID, sItem, iImage);
	}
	SetItemTextColour(iTempID, wxColour(70, 70, 70));
	return iTempID;
}

void UFiletree::OnItemExpanding(wxTreeEvent &oEvent)
{
	wxString sDir(getFullpath(oEvent.GetItem()));
	
	ReloadDir(oEvent.GetItem(), /*sDir, */false);
}

void UFiletree::OnItemCollapsed(wxTreeEvent &oEvent)
{
	wxString sFullpath(getFullpath(oEvent.GetItem()));
	wxDir oDir(sFullpath);
	
	if (itemExists(oEvent.GetItem()))
	{
		DeleteChildren(oEvent.GetItem());
		if (oDir.HasFiles() || oDir.HasSubDirs())
		{
			SetItemHasChildren(oEvent.GetItem(), true);
		}
	}
}

void UFiletree::OnItemActivated(wxTreeEvent &oEvent)
{
	Activate(oEvent.GetItem());
}

void UFiletree::OnItemBeginRename(wxTreeEvent &oEvent)
{
	sTempItemLabel = GetItemText(oEvent.GetItem());
}

void UFiletree::OnItemEndRename(wxTreeEvent &oEvent)
{
	if (!Rename(oEvent.GetItem(), oEvent.GetLabel()))
	{
		SetItemText(oEvent.GetItem(), sTempItemLabel);
	}
}

void UFiletree::ReloadDir(wxTreeItemId iItemID, /*const wxString &sOptDir, */bool bCheckOnly)
{
	/* I don't know if we need this.
	 * I was trying to reproduce a bug when I wrote this
	 * but it sounds like a good idea anyway, so I just commented it and had it for keeps.
	if (!UFileHandler::isDir(sOptDir.mb_str()))
	{
		return;
	}
	*/
	
	//wxString sDir(sOptDir != _("") ? sOptDir : getFullpath(iItemID));
	wxString sDir(getFullpath(iItemID));
	wxDir oDir(sDir);
	wxString sItem;
	bool bCont = oDir.GetFirst(&sItem);
	if (GetChildrenCount(iItemID) == 0 && bCont)
	{
		SetItemHasChildren(iItemID, true);
		if (bCheckOnly)
		{
			return;
		}
	}
	while (bCont)
	{
		if ((sItem.Left(1) != _(".") || bSeeHidden) && !itemExists(sDir + _("/") + sItem))
		{
			addItem(iItemID, sItem);
		}
		bCont = oDir.GetNext(&sItem);
	}
	
	sItem = _("");
	std::vector<wxTreeItemId> iDeletedItems;
	wxTreeItemId iCurrent = GetRootItem();
	while (iCurrent = getNextID(iCurrent))
	{
		sItem = getFullpath(iCurrent);
		if (!wxDir::Exists(sItem) && !wxFile::Exists(sItem))
		{
			iDeletedItems.push_back(iCurrent);
		}
	}
	for (unsigned int i = 0; i < iDeletedItems.size(); i++)
	{
		if (getSingleTreeSelection() == iDeletedItems[i])
		{
			wxTreeItemId iParent = GetItemParent(iDeletedItems[i]);
			wxString sParentDir(getFullpath(iParent));
			
			wxTreeItemId iNext = GetNextSibling(iDeletedItems[i]);
			wxTreeItemId iPrev = GetPrevSibling(iDeletedItems[i]);
			
			if (itemExists(iNext))
			{
				selectTreeItem(iNext);
			}
			else if (itemExists(iPrev))
			{
				selectTreeItem(iPrev);
			}
			else
			{
				selectTreeItem(iParent);
			}
			
			wxDir oDir(sParentDir);
			if (!oDir.HasFiles() && !oDir.HasSubDirs())
			{
				Collapse(iParent);
				SetItemHasChildren(iParent, false);
			}
		}
		if (itemExists(iDeletedItems[i]))
		{
			Delete(iDeletedItems[i]);
		}
	}
	
	SortChildren(iItemID);
}

void UFiletree::SetRoot(const wxString &sDir, bool bSkipLocationText)
{
	if (GetRootItem())
	{
		if (sDir == GetItemText(GetRootItem()) + _("/") || sDir + _("/") == GetItemText(GetRootItem()))
		{
			return;
		}
	}
	
	DeleteAllItems();
	if (!wxFile::Access(sDir, wxFile::read))
	{
		AddRoot(sDir, 3);
	}
	else
	{
		AddRoot(sDir, 1);
		SetItemTextColour(GetRootItem(), wxColour(70, 70, 70));
		SetItemHasChildren(GetRootItem(), true);
		Expand(GetRootItem());
	}
	
	// We do this only to avoid a neverending while-loop,
	// because when we change the value of txtLocation,
	// OnLocationChange() is called, which calls SetRoot
	// again. Which is bad.
	if (!bSkipLocationText)
	{
		txtLocation->SetValue(sDir);
	}
	
	oParentFrame->SetTitle(sDir);
	wxSetWorkingDirectory(sDir);
}

void UFiletree::Activate(wxTreeItemId iItemID)
{
	wxString sItem = getFullpath(iItemID);
	Activate(sItem);
}

void UFiletree::Activate(wxString sItem)
{
	if (wxDir::Exists(sItem))
	{
		sItem.Replace(_("//"), _("/"));
		SetRoot(sItem, false);
	}
	else
	{
		if (UFileHandler::isExecutable((const char *)sItem.mb_str()))
		{
			wxString sRunString;
			sRunString = sItem + _("&");
			system((const char *)sRunString.mb_str());
		}
		else
		{
			wxString sEnding;
			wxString sApp;
			sEnding = (wxChar *)oSettings->getEnding((const char *)sItem.mb_str()).c_str();
			if (oSettings->exists((const char *)sEnding.mb_str()))
			{
				sApp = (wxChar *)oSettings->getRunString((const char *)sEnding.mb_str()).c_str();
			}
			else
			{
				sApp = (wxChar *)oSettings->getGeneral("DEFAULT_APP").c_str();
			}
			if (sApp != _(""))
			{
				wxString sRunString;
				sRunString = wxString::Format(_("%s \"%s\""), (const char *)sApp.mb_str(), (const char *)sItem.mb_str());
				if (oSettings->getTerminal((const char *)sEnding.mb_str()))
				{
					run(sRunString, true);
				}
				else
				{
					system((const char *)wxString(sRunString + _(" &")).mb_str());
				}
			}
		}
	}
}

bool UFiletree::addFile()
{
	if (!somethingSelected())
	{
		return false;
	}
	
	wxTreeItemId iItemID = getSingleTreeSelection();
	wxString sFilename;
	wxString sRunString;
	wxString sParentDir;
	wxString sAllegedDir;
	wxTreeItemId iParentID;
	int iResult;
	sFilename = wxGetTextFromUser(_("Enter name of new file"), _("Name"), _(""));
	
	if (sFilename == _(""))
	{
		return false;
	}
	
	sAllegedDir = getFullpath(iItemID);
	if (wxDir::Exists(sAllegedDir))
	{
		iParentID = iItemID;
		sParentDir = sAllegedDir;
	}
	else
	{
		iParentID = GetItemParent(iItemID);
		sParentDir = getFullpath(iParentID);
	}
	if (!wxFile::Exists(sParentDir + _("/") + sFilename))
	{
		sRunString = _("touch \"") + sParentDir + _("/") + sFilename + _("\"");
		iResult = system((const char *)sRunString.mb_str());
		if (iResult == 0)
		{
			return true;
		}
	}
	return false;
}

bool UFiletree::addFolder()
{
	if (!somethingSelected())
	{
		return false;
	}
	
	wxTreeItemId iItemID = getSingleTreeSelection();
	wxString sFoldername;
	wxString sRunString;
	wxString sParentDir;
	wxString sAllegedDir;
	wxTreeItemId iParentID;
	int iResult;
	sFoldername = wxGetTextFromUser(_("Enter name of new folder"), _("Name"), _(""));
	
	if (sFoldername == _(""))
	{
		return false;
	}
	
	sAllegedDir = getFullpath(iItemID);
	if (wxDir::Exists(sAllegedDir))
	{
		iParentID = iItemID;
		sParentDir = sAllegedDir;
	}
	else
	{
		iParentID = GetItemParent(iItemID);
		sParentDir = getFullpath(iParentID);
	}
	if (!wxDir::Exists(sParentDir + _("/") + sFoldername))
	{
		sRunString = _("mkdir \"") + sParentDir + _("/") + sFoldername + _("\"");
		iResult = system((const char *)sRunString.mb_str());
		if (iResult == 0)
		{
			return true;
		}
	}
	return false;
}

bool UFiletree::Rename(wxTreeItemId iItemID, const wxString &sNewName)
{
	wxString sRunString;
	wxString sItem;
	wxString sNewItem;
	bool bResult = false;
	sItem = getFullpath(iItemID);
	sNewItem = sItem.Left(sItem.Length() - GetItemText(iItemID).Length()) + sNewName;
	if (sNewName.Length() > 0)
	{
		bResult = UFileHandler::move((const char *)sItem.mb_str(), (const char *)sNewItem.mb_str());
	}
	return bResult;
}

void UFiletree::GoUp()
{
	wxString sOldPath = txtLocation->GetValue();
	wxString sNewPath;
	
	if (sOldPath.Right(1) == _("/"))
	{
		sOldPath = sOldPath.Left(sOldPath.Length() - 1);
	}
	
	sNewPath = sOldPath.Left(sOldPath.Find('/', true));
	
	if (sNewPath == _(""))
	{
		sNewPath = _("/");
	}

	if (sNewPath == sOldPath + _("/"))
	{
		return;
	}
	
	SetRoot(sNewPath, false);
	SetFocus();
}

void UFiletree::rightPress()
{
	wxTreeItemIdValue iDummy = 0;
	wxString sFullpath;
	wxTreeItemId iItem;
	if (!somethingSelected())
	{
		return;
	}
	iItem = getSingleTreeSelection();
	sFullpath = getFullpath(iItem);
	if (!wxDir::Exists(sFullpath))
	{
		return;
	}
	wxDir oDir(sFullpath);
	if (oDir.HasFiles() || oDir.HasSubDirs())
	{
		if (IsExpanded(iItem))
		{
			selectTreeItem(GetFirstChild(iItem, iDummy));
		}
		else
		{
			Expand(iItem);
		}
	}
}

void UFiletree::leftPress()
{
	if (!somethingSelected())
	{
		return;
	}
	
	if (IsExpanded(getSingleTreeSelection()))
	{
		CollapseAndReset(getSingleTreeSelection());
	}
	else if (GetItemParent(getSingleTreeSelection()))
	{
		selectTreeItem(GetItemParent(getSingleTreeSelection()));
	}
}

void UFiletree::selectNextItem()
{
	wxTreeItemId iItem;
	if (!somethingSelected())
	{
		return;
	}
	iItem = getSingleTreeSelection();
	if (getNextID(iItem))
	{
		selectTreeItem(getNextID(getSingleTreeSelection()));
	}
}

void UFiletree::selectTreeItem(wxTreeItemId iItem)
{
	SelectItem(iItem);
}

void UFiletree::pressHome()
{
	selectTreeItem(GetRootItem());
}

void UFiletree::pressEnd()
{
	selectTreeItem(GetLastChild(GetRootItem()));
}

void UFiletree::openWith()
{
	wxTreeItemId iItem;
	if (iItem = getSingleTreeSelection())
	{
		wxString sRunString;
		wxString sFullpath;
		sRunString = wxGetTextFromUser(_("Run with..."), _("Runstring:"), _("gvim '%s'"));
		if (sRunString.Find(_("%s")) == -1)
		{
			sRunString += _(" '%s'");
		}
		
		sFullpath = getFullpath(iItem);
		sRunString.Replace(_("%s"), sFullpath);
		wxExecute(sRunString);
	}
}

bool UFiletree::somethingSelected()
{
	wxArrayTreeItemIds iSelections;
	return (GetSelections(iSelections) > 0);
}

void UFiletree::edit()
{
	if (!somethingSelected())
	{
		return;
	}
	EditLabel(getSingleTreeSelection());
}

bool UFiletree::itemExists(const wxTreeItemId &iItem)
{
	wxTreeItemId iCurrent = GetRootItem();
	
	if (iCurrent == iItem)
	{
		return true;
	}
	
	bool bReturn = false;
	while (iCurrent = getNextID(iCurrent))
	{
		if (iCurrent == iItem)
		{
			bReturn = true;
			break;
		}
	}
	return bReturn;
}

bool UFiletree::itemExists(const wxString &sItem)
{
	wxTreeItemId iCurrent = GetRootItem();
	
	if (GetItemText(iCurrent) == sItem)
	{
		return true;
	}
	
	bool bReturn = false;
	while (iCurrent = getNextID(iCurrent))
	{
		if (getFullpath(iCurrent) == sItem)
		{
			bReturn = true;
			break;
		}
	}
	return bReturn;
}

// We'll do this until we get a real GetNextItem() function in wxWindows,
// one that simply, and only, fucking, gets the next item in the list, regardless of parenthood.
wxTreeItemId UFiletree::getNextID(const wxTreeItemId &iItem)
{
	wxTreeItemId iNext;
	wxTreeItemId iCurrent;
	if (IsExpanded(iItem) && GetChildrenCount(iItem) > 0)
	{
		wxTreeItemIdValue iDummy;
		return GetFirstChild(iItem, iDummy);
	}
	
	iCurrent = iItem;
	
	if (iNext = GetNextSibling(iCurrent))
	{
		return iNext;
	}
	
	while (GetItemParent(iCurrent))
	{
		iCurrent = GetItemParent(iCurrent);
		if (iNext = GetNextSibling(iCurrent))
		{
			return iNext;
		}
	}
	
	// And finally... we bail out like chickenshit, by knowingly returning something that isn't valid.
	// Although actually... it should never really come to this.
	// The while-loop above should always return iNext at some point.
	return GetNextSibling(iItem);
}

wxTreeItemId UFiletree::getPrevID(const wxTreeItemId &iItem)
{
	wxTreeItemId iCurrent = GetRootItem();
	
	if (iItem == iCurrent)
	{
		return iItem;
	}
	
	while (iItem != getNextID(iCurrent))
	{
		iCurrent = getNextID(iCurrent);
	}
	return iCurrent;
}

std::vector<wxTreeItemId> UFiletree::getTreeSelections()
{
	wxArrayTreeItemIds iSelections;
	GetSelections(iSelections);
	
	std::vector<wxTreeItemId> iNewSelections;
	for (unsigned int i = 0; i < iSelections.GetCount(); i++)
	{
		iNewSelections.push_back(iSelections[i]);
	}
	return iNewSelections;
}

wxTreeItemId UFiletree::getSingleTreeSelection()
{
	wxArrayTreeItemIds iSelections;
	GetSelections(iSelections);
	if (iSelections.GetCount() > 0)
	{
		return iSelections[0];
	}
	wxTreeItemId iSpareItem;
	return iSpareItem;
}

void UFiletree::run(const wxString &sCommand, bool bTerminal)
{
	oParentFrame->oProcessBox->setCommand(sCommand);
	
	if (!oParentFrame->oProcessBox->IsShown())
	{
		oParentFrame->oProcessBox->Show(true);
		oParentFrame->btnProcessClose->Show(true);
		oParentFrame->oMainSizer->Add(oParentFrame->oProcessBox, 0, wxGROW);
		oParentFrame->oMainSizer->Add(oParentFrame->btnProcessClose, 0, wxALIGN_CENTER, wxALIGN_CENTER);
		oParentFrame->oMainSizer->SetItemMinSize(oParentFrame->oProcessBox, 10, 120);
		oParentFrame->oMainSizer->Layout();
		oParentFrame->btnProcessClose->SetFocus();
	}
	
	oParentFrame->oProcessBox->run();
}

